/*
 * Jitsi, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jitsi.impl.neomedia.notify;

import java.beans.*;
import java.net.*;
import java.util.*;

import javax.media.*;

import org.jitsi.impl.neomedia.*;
import org.jitsi.impl.neomedia.device.*;
import org.jitsi.service.audionotifier.*;
import org.jitsi.service.libjitsi.*;
import org.jitsi.service.resources.*;

/**
 * The implementation of the AudioNotifierService.
 *
 * @author Yana Stamcheva
 */
public class AudioNotifierServiceImpl
    implements AudioNotifierService,
               PropertyChangeListener
{
    /**
     * Map of different audio clips.
     */
    private static final Map<AudioClipsKey, AbstractSCAudioClip> audioClips
        = new HashMap<AudioClipsKey, AbstractSCAudioClip>();

    /**
     * If the sound is currently disabled.
     */
    private boolean isMute;

    /**
     * Device config to look for notify device.
     */
    private final DeviceConfiguration deviceConfiguration;

    /**
     * Initializes a new <tt>AudioNotifierServiceImpl</tt> instance.
     */
    public AudioNotifierServiceImpl()
    {
        this.deviceConfiguration
            = NeomediaServiceUtils
                .getMediaServiceImpl()
                    .getDeviceConfiguration();

        deviceConfiguration.addPropertyChangeListener(this);
    }

    /**
     * Creates an SCAudioClip from the given URI and adds it to the list of
     * available audio-s. Uses notification device if any.
     *
     * @param uri the path where the audio file could be found
     * @return a newly created <tt>SCAudioClip</tt> from <tt>uri</tt>
     */
    public AbstractSCAudioClip createAudio(String uri)
    {
        return createAudio(uri, false);
    }

    /**
     * Creates an SCAudioClip from the given URI and adds it to the list of
     * available audio-s.
     *
     * @param uri the path where the audio file could be found
     * @param playback use or not the playback device.
     * @return a newly created <tt>SCAudioClip</tt> from <tt>uri</tt>
     */
    public AbstractSCAudioClip createAudio(String uri, boolean playback)
    {
        AbstractSCAudioClip audioClip;

        synchronized (audioClips)
        {
            AudioClipsKey key = new AudioClipsKey(uri, playback);
            if(audioClips.containsKey(key))
            {
                audioClip = audioClips.get(key);
            }
            else
            {
                ResourceManagementService resources
                    = LibJitsi.getResourceManagementService();
                URL url
                    = (resources == null)
                        ? null
                        : resources.getSoundURLForPath(uri);

                if (url == null)
                {
                    // Not found by the class loader. Perhaps it's a local file.
                    try
                    {
                        url = new URL(uri);
                    }
                    catch (MalformedURLException e)
                    {
                        //logger.error("The given uri could not be parsed.", e);
                        return null;
                    }
                }

                try
                {
                    AudioSystem audioSystem
                        = getDeviceConfiguration().getAudioSystem();

                    if (audioSystem == null)
                        audioClip = new JavaSoundClipImpl(url, this);
                    else if (NoneAudioSystem.LOCATOR_PROTOCOL.equalsIgnoreCase(
                            audioSystem.getLocatorProtocol()))
                        audioClip = null;
                    else
                    {
                        audioClip
                            = new AudioSystemClipImpl(
                                    url,
                                    this, audioSystem,
                                    playback);
                    }
                }
                catch (Throwable t)
                {
                    if (t instanceof ThreadDeath)
                        throw (ThreadDeath) t;
                    // Cannot create audio to play
                    return null;
                }

                audioClips.put(key, audioClip);
            }
        }

        return audioClip;
    }

    /**
     * Removes the given audio from the list of available audio clips.
     *
     * @param audioClip the audio to destroy
     */
    public void destroyAudio(SCAudioClip audioClip)
    {
        synchronized (audioClips)
        {
            AudioClipsKey keyToRemove = null;
            for(Map.Entry<AudioClipsKey, AbstractSCAudioClip> entry
                : audioClips.entrySet())
            {
                if(entry.getValue().equals(audioClip))
                {
                    keyToRemove = entry.getKey();
                    break;
                }
            }

            audioClips.remove(keyToRemove);
        }
    }

    /**
     * Enables or disables the sound in the application. If FALSE, we try to
     * restore all looping sounds if any.
     *
     * @param isMute when TRUE disables the sound, otherwise enables the sound.
     */
    public void setMute(boolean isMute)
    {
        this.isMute = isMute;

        for (AbstractSCAudioClip audioClip : audioClips.values())
        {
            if (isMute)
            {
                audioClip.internalStop();
            }
            else if (audioClip.isLooping())
            {
                // TODO Auto-generated method stub
            }
        }
    }

    /**
     * Returns TRUE if the sound is currently disabled, FALSE otherwise.
     * @return TRUE if the sound is currently disabled, FALSE otherwise
     */
    public boolean isMute()
    {
        return isMute;
    }

    /**
     * The device configuration.
     *
     * @return the deviceConfiguration
     */
    public DeviceConfiguration getDeviceConfiguration()
    {
        return deviceConfiguration;
    }

    /**
     * Listens for changes in notify device
     * @param event the event that notify device has changed.
     */
    public void propertyChange(PropertyChangeEvent event)
    {
        if (DeviceConfiguration.AUDIO_NOTIFY_DEVICE.equals(
                event.getPropertyName()))
        {
            audioClips.clear();
        }
    }

    /**
     * Checks whether the playback and notification configuration
     * share the same device.
     * @return are audio out and notifications using the same device.
     */
    public boolean audioOutAndNotificationsShareSameDevice()
    {
        CaptureDeviceInfo notifyInfo =
            getDeviceConfiguration().getAudioSystem()
                .getDevice(AudioSystem.NOTIFY_INDEX);
        CaptureDeviceInfo playbackInfo =
            getDeviceConfiguration().getAudioSystem()
                .getDevice(AudioSystem.PLAYBACK_INDEX);

        if(notifyInfo != null && playbackInfo != null)
        {
            return notifyInfo.getLocator().equals(playbackInfo.getLocator());
        }

        return false;
    }

    /**
     * Key for clips.
     */
    private static class AudioClipsKey
    {
        /**
         * The uri.
         */
        private String uri;

        /**
         * Is it playback.
         */
        private boolean isPlayback;

        /**
         * Constructs key.
         * @param uri by uri
         * @param playback and playback
         */
        private AudioClipsKey(String uri, boolean playback)
        {
            this.uri = uri;
            isPlayback = playback;
        }

        @Override
        public boolean equals(Object o)
        {
            AudioClipsKey that = (AudioClipsKey) o;

            if(isPlayback != that.isPlayback)
                return false;
            if(uri != null ? !uri.equals(that.uri) : that.uri != null)
                return false;

            return true;
        }
    }
}
